from rest_framework.permissions import BasePermission

class Permissions(BasePermission):
    def __init__(self):
        super().__init__()  # Call parent's init
        self.all_permissions = []  # Initialize the list

    def set_permission(self, permission_class):
        # Create a permission entry with a default "all actions" setting
        perm_entry = PermissionEntry(permission_class, all_actions=True)
        self.all_permissions.append(perm_entry)
        return perm_entry

    def retrieve_permissions(self, view):
        # Apply permissions based on the view's action, considering exclusions
        return [
            perm_entry.permission_class()
            for perm_entry in self.all_permissions
            if (perm_entry.all_actions or view.action in perm_entry.actions) 
            and view.action not in perm_entry.excluded_actions
        ]

class PermissionEntry:
    def __init__(self, permission_class, all_actions=False):
        self.permission_class = permission_class
        self.all_actions = all_actions
        self.actions = []
        self.excluded_actions = []  # New: track excluded actions

    def only(self, *actions):
        # Handle both array input and multiple arguments
        if len(actions) == 1 and isinstance(actions[0], (list, tuple)):
            self.actions.extend(actions[0])
        else:
            self.actions.extend(actions)
        self.all_actions = False  # Turn off all_actions if specific actions are set
        return self  # Allow chaining

    def ignore(self, *actions):  # New method
        # Handle both array input and multiple arguments
        if len(actions) == 1 and isinstance(actions[0], (list, tuple)):
            self.excluded_actions.extend(actions[0])
        else:
            self.excluded_actions.extend(actions)
        return self  # Allow chaining


# celery_config.py

from __future__ import absolute_import, unicode_literals
from django_petra.scheduler.schedule import bind_beat_schedule
import os
from celery import Celery
from django.apps import apps
import importlib

def configure_celery(project_name, type = 'redis'):

    # Set the default Django settings module for the 'celery' program.
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', f'{project_name}.settings')

    # Create a Celery instance.
    app = Celery(project_name)

    # Load task modules from all registered Django app configs.
    app.config_from_object('django.conf:settings', namespace='CELERY')

    app.autodiscover_tasks(lambda: [n.name for n in apps.get_app_configs()])

    app.conf.broker_url = os.environ.get('CELERY_BROKER')

    # Dynamically import schedules from {project_name}.friend_scheduler
    schedules_module = importlib.import_module(f'{project_name}.petra_config.scheduler')
    app.conf.beat_schedule = bind_beat_schedule(schedules=schedules_module.schedules)

    # RedBeat Configuration
    if type == 'redis':
        app.conf.redbeat_redis_url = os.environ.get('CELERY_BROKER')
        # Use RedBeatScheduler
        app.conf.beat_scheduler = 'redbeat.RedBeatScheduler'
    
    return app



import boto3
import mimetypes
from botocore.exceptions import ClientError
from django_petra.env import get_env

# Retrieve environment variables
aws_access_key_id = get_env('AWS_ACCESS_KEY_ID')
aws_secret_access_key = get_env('AWS_SECRET_ACCESS_KEY')
aws_region = get_env('S3_AWS_REGION')
bucket_name = get_env('S3_BUCKET_NAME')

class S3Storage:
    def __init__(self):
        self.s3_bucket = bucket_name
        self.s3_client = boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=aws_region
        )

    def put(self, path, file):
        try:
            file_content = file.read()
            content_type, _ = mimetypes.guess_type(file.name)
            self.s3_client.put_object(
                Body=file_content,
                Bucket=self.s3_bucket,
                Key=path, 
                ContentType=content_type
            )
        except ClientError as e:
            print(f"Error putting object to S3: {e}")

    def get(self, path):
        try:
            response = self.s3_client.get_object(Bucket=self.s3_bucket, Key=path)
            return response['Body'].read()
        except ClientError as e:
            print(f"Error getting object from S3: {e}")

    def delete(self, path):
        try:
            self.s3_client.delete_object(Bucket=self.s3_bucket, Key=path)
        except ClientError as e:
            print(f"Error deleting object from S3: {e}")

    def update(self, path, file):
        self.delete(path)
        self.put(path, file)

    def exists(self, path):
        try:
            self.s3_client.head_object(Bucket=self.s3_bucket, Key=path)
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False
            else:
                print(f"Error checking object existence in S3: {e}")

    def download(self, source_path, destination_path):
        try:
            self.s3_client.download_file(self.s3_bucket, source_path, destination_path)
        except ClientError as e:
            print(f"Error downloading object from S3: {e}")

    def url(self, path):
        try:
            return self.s3_client.generate_presigned_url('get_object', Params={'Bucket': self.s3_bucket, 'Key': path})
        except ClientError as e:
            print(f"Error generating URL for S3 object: {e}")

    def temporary_url(self, path, expiration=3600):
        try:
            return self.s3_client.generate_presigned_url('get_object', Params={'Bucket': self.s3_bucket, 'Key': path}, ExpiresIn=expiration)
        except ClientError as e:
            print(f"Error generating temporary URL for S3 object: {e}")

    def all_files(self, directory):
        try:
            objects = self.s3_client.list_objects(Bucket=self.s3_bucket, Prefix=directory)
            return [obj['Key'] for obj in objects.get('Contents', [])]
        except ClientError as e:
            print(f"Error listing all files in S3 directory: {e}")

    def directories(self, directory):
        try:
            objects = self.s3_client.list_objects(Bucket=self.s3_bucket, Prefix=directory, Delimiter='/')
            return [common_prefix['Prefix'] for common_prefix in objects.get('CommonPrefixes', [])]
        except ClientError as e:
            print(f"Error listing directories in S3 directory: {e}")

    def append(self, path, contents):
        current_contents = self.get(path)
        new_contents = current_contents + contents
        self.update(path, new_contents)

    def prepend(self, path, contents):
        current_contents = self.get(path)
        new_contents = contents + current_contents
        self.update(path, new_contents)

    def copy(self, source, destination):
        self.s3_client.copy_object(Bucket=self.s3_bucket, CopySource=f'/{self.s3_bucket}/{source}', Key=destination)

    def move(self, source, destination):
        self.copy(source, destination)
        self.delete(source)

    def content_type(self, path):
        content_type, _ = mimetypes.guess_type(path)
        if not content_type:
            content_type = 'application/octet-stream'
        return content_type



from dotenv import load_dotenv
import os

def load_env():
    """Load environment variables from .env file."""
    load_dotenv()

def get_env(key: str, default=None) -> str:
    """
    Retrieve environment variable value.
    
    Args:
        key: The environment variable key
        default: Default value if key is not found
        
    Returns:
        The environment variable value or default if not found
    """
    return os.getenv(key, default)


from pathlib import Path
from importlib import import_module
from django.urls import path, include
from .path import modules_path

def bind_modules_app():
    the_modules_path = Path(modules_path()).resolve()
    
    # Initialize an empty list to store installed apps
    INSTALLED_APPS = [
        # third party apps
        'rest_framework',
        'rest_framework.authtoken',
        'celery',
    ]
    
    INSTALLED_APPS.append('django_petra.project')
    INSTALLED_APPS.append('django_petra.cors')

    # Loop through each module directory
    for module_dir in the_modules_path.iterdir():
        if module_dir.is_dir():
            # Construct the path to the urls.py file in the module directory
            module_urls_path = module_dir / 'urls.py'
            
            # Get the name of the module
            module_name = module_dir.name
            
            # Check if the urls.py file exists
            if module_urls_path.exists():
                # Append the module to INSTALLED_APPS
                INSTALLED_APPS.append(f'modules.{module_name}')

    # Return the INSTALLED_APPS list
    return INSTALLED_APPS

def bind_modules_urls():
    the_modules_path = Path(modules_path()).resolve()
    
    # Initialize an empty list to store urlpatterns
    urlpatterns = []
    urlpatterns.append(path(f'hello-django-petra/', include(('django_petra.project.urls', 'django_petra.project'))))

    # Loop through each module directory
    for module_dir in the_modules_path.iterdir():
        if module_dir.is_dir():
            # Get the name of the module
            module_name = module_dir.name
            
            # Construct the path to the urls.py file in the module directory
            module_urls_path = module_dir / 'urls.py'
            
            # Check if the urls.py file exists
            if module_urls_path.exists():
                # Construct the import path for the module's urls
                module_urls = f'modules.{module_name}.urls'

                # Import the module to get its version
                urls_module = import_module(module_urls)
                # Get version from module, default to 'v1' if not specified
                version = getattr(urls_module, 'API_VERSION', 'v1')
                
                # Append a path to urlpatterns, including the module's urls
                urlpatterns.append(path(f'api/{version}/{module_name}/', include((module_urls, module_name))))

    # Return the urlpatterns list
    return urlpatterns

def show_modules_url():
    from django_petra.router import Router, Route
    the_modules_path = Path(modules_path()).resolve()
    
    # Initialize an empty list to store urlpatterns
    the_routes = []
    # urlpatterns.append(path(f'hello-django-petra/', include(('django_petra.project.urls', 'django_petra.project'))))

    # Loop through each module directory
    for module_dir in the_modules_path.iterdir():
        if module_dir.is_dir():
            # Get the name of the module
            module_name = module_dir.name
            
            # Construct the path to the urls.py file in the module directory
            module_urls_path = module_dir / 'urls.py'
            
            # Check if the urls.py file exists
            if module_urls_path.exists():
                # Construct the import path for the module's urls
                module_urls = f'modules.{module_name}.urls'
                
                 # Dynamically import the module's urls
                urls_module = import_module(module_urls)
                
                # Get the routes from the imported module
                module_routes = getattr(urls_module, 'routes', [])
                
                generate_routes = [(route) for route in Route(module_routes).show_lists()]
                
                # Append a path to urlpatterns, including the module's urls
                the_routes.extend(generate_routes)

    # Return the urlpatterns list
    return the_routes

from django_petra.initialize import init_django_petra

def initialize():
    init_django_petra()
    





from rest_framework import serializers
from uuid import UUID
from datetime import datetime
from collections import OrderedDict

class RawQuerySerializer(serializers.Serializer):
    def generate_fields(self, data):
        get_fields_ordered = list(OrderedDict.fromkeys(self.get_fields()))  # Fields specified in 'get_fields'
        meta_exclude = self.get_meta_exclude()  # Fields to be excluded from Meta

        if not data:
            return []

        # Get the fields from the 'Meta' class or from the first data dictionary
        model_fields = self.get_meta_fields() or list(data[0].keys())

        for model_field in model_fields:
            # if model_field not in list(data[0].keys()):
            #     raise serializers.ValidationError(f"Field '{model_field}' is not present in the provided data.")

            if model_field not in get_fields_ordered and model_field not in meta_exclude:
                # Determine the field type based on the data type in the first dictionary
                field_type = serializers.CharField()

                # Mapping between Python types and Django model field types
                type_mapping = {
                    int: serializers.IntegerField,
                    float: serializers.FloatField,
                    str: serializers.CharField,
                    bool: serializers.BooleanField,
                    UUID: serializers.UUIDField,
                    datetime: serializers.DateTimeField,
                    list: serializers.ListField,
                    dict: serializers.DictField,
                    # Add more mappings as needed
                }

                python_type = type(data[0][model_field])
                if python_type in type_mapping:
                    field_type = type_mapping[python_type]()

                self.fields[model_field] = field_type
                
        # Remove fields that are not in get_fields_ordered and not in meta_exclude
        for get_field in get_fields_ordered:
            if get_field not in model_fields and get_field not in meta_exclude:
                del self.fields[get_field]

        # Check for fields in meta_exclude that are not in the data
        for except_field in meta_exclude:
            if except_field not in model_fields:
                raise serializers.ValidationError(f"Field '{except_field}' is listed in meta_exclude but not present in the provided data.")

        return model_fields

    def __init__(self, instance=None, data=None, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)
        self.generate_fields(data)

    def get_meta_fields(self):
        # Check if the 'Meta' attribute exists in the class and has the 'fields' attribute
        if hasattr(self, 'Meta') and hasattr(self.Meta, 'fields'):
            return self.Meta.fields
        else:
            return None

    def get_meta_exclude(self):
        # Check if the 'Meta' attribute exists in the class and has the 'exclude' attribute
        if hasattr(self, 'Meta') and hasattr(self.Meta, 'exclude'):
            return self.Meta.exclude
        else:
            return []

    def the_update_fields(self, instance):
        updated_values = {}
        if hasattr(self, 'update_fields') and callable(getattr(self, 'update_fields', None)):
            if instance is not None:
                updated_fields = self.update_fields(getField=lambda field_name, default=None: instance.get(field_name, default))
                for field, updated_value in OrderedDict(updated_fields).items():
                    updated_values[field] = updated_value
        return updated_values
    
    def the_bind_relations(self, instance):
        updated_values = {}
        if hasattr(self, 'bind_relations') and callable(getattr(self, 'bind_relations', None)):
            if instance is not None:
                updated_fields = self.bind_relations(getField=lambda field_name, default=None: instance.get(field_name, default))
                for field, updated_value in OrderedDict(updated_fields).items():
                    updated_values[field] = updated_value
        return updated_values
    
    
    def to_representation(self, instance):
        representation = super().to_representation(instance)
        updated_fields = self.the_update_fields(instance)
        
        for field, value in OrderedDict(updated_fields).items():
            representation[field] = value
            
        for field, value in OrderedDict(self.the_bind_relations(instance)).items():
            representation[field] = value
            
        return representation



from django.db import models

class SoftDeleteManager(models.Manager):
    def save_soft_delete(self):
        self.is_deleted = True
        self.save()
        return True

    def get_soft_delete(self):
        return self.filter(is_deleted=True)

    def get_unsoft_delete(self):
        return self.filter(is_deleted=False)


import importlib
from django_petra.path import base_module_name

the_base_module_name = base_module_name()

# Import CORS settings dynamically
cors_module = importlib.import_module(f"{the_base_module_name}.petra_config.cors")
ALLOWED_METHODS = cors_module.ALLOWED_METHODS
ALLOWED_ORIGINS = cors_module.ALLOWED_ORIGINS
ALLOWED_ORIGINS_PATTERNS = cors_module.ALLOWED_ORIGINS_PATTERNS
ALLOWED_HEADERS = cors_module.ALLOWED_HEADERS
EXPOSED_HEADERS = cors_module.EXPOSED_HEADERS
MAX_AGE = cors_module.MAX_AGE
SUPPORTS_CREDENTIALS = cors_module.SUPPORTS_CREDENTIALS
CORS_ALLOW_ALL_ORIGINS = cors_module.CORS_ALLOW_ALL_ORIGINS

class Settings:
    """
    Shadow Django's settings with a little logic
    """

    @property
    def CORS_ALLOW_HEADERS(self):
        return ALLOWED_HEADERS

    @property
    def CORS_ALLOW_METHODS(self):
        return ALLOWED_METHODS

    @property
    def CORS_ALLOW_CREDENTIALS(self):
        return SUPPORTS_CREDENTIALS

    @property
    def CORS_ALLOW_PRIVATE_NETWORK(self):
        return False

    @property
    def CORS_PREFLIGHT_MAX_AGE(self):
        return MAX_AGE

    @property
    def CORS_ALLOW_ALL_ORIGINS(self):
        return CORS_ALLOW_ALL_ORIGINS

    @property
    def CORS_ALLOWED_ORIGINS(self):
        return tuple(ALLOWED_ORIGINS)

    @property
    def CORS_ALLOWED_ORIGIN_REGEXES(self):
        return ALLOWED_ORIGINS_PATTERNS

    @property
    def CORS_EXPOSE_HEADERS(self):
        return EXPOSED_HEADERS

    @property
    def CORS_URLS_REGEX(self):
        return r"^.*$"

conf = Settings()



from __future__ import annotations
from django.apps import AppConfig
from django.core.checks import register
from django.core.checks import Tags
from django_petra.cors.checks import check_settings

class CorsAppConfig(AppConfig):
    name = 'django_petra.cors'

    def ready(self) -> None:
        register(Tags.security)(check_settings)


import os
import shutil
import mimetypes

from django_petra.path import base_path

class LocalStorage:
    def __init__(self):
        self.base_path = base_path()

    def get_full_path(self, path):
        return os.path.join(self.base_path, 'storage', 'public', path)

    def put(self, path, file):
        full_path = self.get_full_path(path)
        file_content = file.read()
        os.makedirs(os.path.dirname(full_path), exist_ok=True)
        
        # Check if contents is bytes or string
        mode = 'wb' if isinstance(file_content, bytes) else 'w'
        with open(full_path, mode) as file:
            file.write(file_content)

    def get(self, path):
        full_path = self.get_full_path(path)
        with open(full_path, 'rb') as file:
            return file.read()

    def delete(self, path):
        full_path = self.get_full_path(path)
        try:
            os.remove(full_path)
        except FileNotFoundError as e:
            print(f"Error deleting local file: {e}")

    def update(self, path, file):
        self.delete(path)
        self.put(path, file)

    def exists(self, path):
        full_path = self.get_full_path(path)
        return os.path.exists(full_path)

    def download(self, source_path, destination_path):
        full_source_path = self.get_full_path(source_path)
        try:
            shutil.copy(full_source_path, destination_path)
        except FileNotFoundError as e:
            print(f"Error downloading local file: {e}")

    def url(self, path):
        full_path = self.get_full_path(path)
        return f"file://{os.path.abspath(full_path)}"

    def all_files(self, directory):
        full_directory = self.get_full_path(directory)
        try:
            return [os.path.join(full_directory, f) for f in os.listdir(full_directory) if os.path.isfile(os.path.join(full_directory, f))]
        except Exception as e:
            print(f"Error listing all files in local directory: {e}")

    def directories(self, directory):
        full_directory = self.get_full_path(directory)
        try:
            return [d for d in os.listdir(full_directory) if os.path.isdir(os.path.join(full_directory, d))]
        except Exception as e:
            print(f"Error listing directories in local directory: {e}")

    def append(self, path, contents):
        full_path = self.get_full_path(path)
        try:
            with open(full_path, 'a') as file:
                file.write(contents)
        except Exception as e:
            print(f"Error appending to local file: {e}")

    def prepend(self, path, contents):
        full_path = self.get_full_path(path)
        try:
            with open(full_path, 'r') as file:
                old_contents = file.read()
            with open(full_path, 'w') as file:
                file.write(contents + old_contents)
        except Exception as e:
            print(f"Error prepending to local file: {e}")

    def copy(self, source, destination):
        full_source_path = self.get_full_path(source)
        full_destination_path = self.get_full_path(destination)
        try:
            shutil.copy(full_source_path, full_destination_path)
        except Exception as e:
            print(f"Error copying local file: {e}")

    def move(self, source, destination):
        full_source_path = self.get_full_path(source)
        full_destination_path = self.get_full_path(destination)
        try:
            shutil.move(full_source_path, full_destination_path)
        except Exception as e:
            print(f"Error moving local file: {e}")

    def content_type(self, path):
        content_type, _ = mimetypes.guess_type(path)
        if not content_type:
            content_type = 'application/octet-stream'
        return content_type



from django_petra.initialize.initialize import (
  init_cors_middleware,
  init_all_modules,
  init_module_urls,
)

def init_django_petra():
  init_cors_middleware()
  init_all_modules()
  init_module_urls()


from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from email.mime.image import MIMEImage
from django_petra.mailer.inline_css import inline_css

class SendMail:
    """
    A utility class for sending emails using Django's EmailMessage.

    Usage:
    mail = SendMail()
    mail.from_email("sender@example.com").to(["recipient@example.com"]).subject("Subject").html_body("<p>HTML Content</p>").send()
    """

    def __init__(self):
        """
        Initialize the SendMail object with an EmailMessage instance.
        """
        self.email = EmailMultiAlternatives()

    def from_email(self, sender):
        """
        Set the sender email address.

        Args:
            sender (str): The email address of the sender.

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.from_email = sender
        return self

    def to(self, recipients):
        """
        Set the recipient email addresses.

        Args:
            recipients (str or list): The email address or list of email addresses of the recipients.

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.to = recipients if isinstance(recipients, list) else [recipients]
        return self

    def cc(self, cc_recipients):
        """
        Set the carbon copy (cc) recipient email addresses.

        Args:
            cc_recipients (str or list): The email address or list of email addresses of the cc recipients.

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.cc = cc_recipients if isinstance(cc_recipients, list) else [cc_recipients]
        return self

    def bcc(self, bcc_recipients):
        """
        Set the blind carbon copy (bcc) recipient email addresses.

        Args:
            bcc_recipients (str or list): The email address or list of email addresses of the bcc recipients.

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.bcc = bcc_recipients if isinstance(bcc_recipients, list) else [bcc_recipients]
        return self

    def subject(self, subject):
        """
        Set the email subject.

        Args:
            subject (str): The subject of the email.

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.subject = subject
        return self

    def body(self, body, content_type='text/plain'):
        """
        Set the email body.

        Args:
            body (str): The body content of the email.
            content_type (str): The content type of the email body (default is 'text/plain').

        Returns:
            SendMail: The current SendMail instance.
        """
        self.email.attach_alternative(body, content_type)
        return self

    def html_body(self, html_body):
        """
        Set the email body as HTML content.

        Args:
            html_body (str): The HTML content of the email body.

        Returns:
            SendMail: The current SendMail instance.
        """
        return self.body(html_body, content_type='text/html')

    def template(self, template_name, context):
        """
        Set the email body using a Django template.

        Args:
            template_name (str): The name of the template file (without extension).
            context (dict): The context data to render the template.

        Returns:
            SendMail: The current SendMail instance.
        """
        html_body = inline_css(render_to_string(f"mail/{template_name}.html", context))
        return self.html_body(html_body)

    def attach(self, file_paths):
        """
        Attach files to the email.

        Args:
            file_paths (list): List of file paths to be attached.

        Returns:
            SendMail: The current SendMail instance.
        """
        for file_path in file_paths:
            with open(file_path, 'rb') as file:
                self.email.attach(file.name, file.read())
        return self

    def attach_image(self, image_path):
        """
        Attach an image to the email as an inline image.

        Args:
            image_path (str): The path to the image file.

        Returns:
            SendMail: The current SendMail instance.
        """
        with open(image_path, 'rb') as image:
            img = MIMEImage(image.read())
            img.add_header('Content-ID', '<{}>'.format(image.name))
            self.email.attach(img)
        return self

    def send(self):
        """
        Send the email using the configured EmailMessage instance.

        Returns:
            None
        """
        
        self.email.send()



import os
from collections import OrderedDict
from django_petra.path import base_path
from django_petra.env import get_env
    
def load_sql(sql_file, load_from_module = False):
    if load_from_module == True:
      components = sql_file.split('.')
      path = base_path('modules', get_env('RAW_SQL_MODULE_DIR', 'sql_quries'), *components)
      path_with_extension = path + ".sql"
    else:
      components = sql_file.split('.')
      path = base_path(get_env('RAW_SQL_DIR', 'raw_sql'), *components)
      path_with_extension = path + ".sql"

    # Open and read the SQL file
    with open(path_with_extension, 'r') as file:
        return file.read()
      
def fetch_all_to_dictionary(cursor):
    columns = [col[0] for col in cursor.description]
    return [dict(zip(columns, row)) for row in cursor.fetchall()]


def raw_query_collection(request, results, wrap="data", type='paginate', serializer_class=None):
    """
    Paginate a raw query result and construct a Laravel-style response.

    Parameters:
    - request: The Django request object containing query parameters.
    - results: The list of results to paginate.
    - wrap: The key to wrap around the paginated results in the response.

    Query Parameters:
    - page: The page number to retrieve (default: 1).
    - per_page: The number of items per page (default: 10).

    Returns:
    An OrderedDict containing paginated results and pagination information.
    """

    # Get the page and per_page values from the request query parameters
    page = int(request.query_params.get('page', 1))
    per_page = int(request.query_params.get('per_page', 10))

    # Calculate indices for the range of items to be displayed on the current page
    from_index = (page - 1) * per_page + 1
    to_index = min(page * per_page, len(results))

    # Calculate start and end indices for slicing the paginated results
    start_index = (page - 1) * per_page
    end_index = page * per_page

    # Slice the results to get the paginated subset
    paginated_results = results[start_index:end_index]

    # Calculate the total number of pages
    total_pages = (len(results) + per_page - 1) // per_page

    # Determine the next and previous page numbers
    next_page = page + 1 if page < total_pages else None
    prev_page = page - 1 if page > 1 else None

    # Build URLs for the next and previous pages
    base_url = request.build_absolute_uri(request.path)
    next_page_url = f"{base_url}?page={next_page}&per_page={per_page}" if next_page else None
    prev_page_url = f"{base_url}?page={prev_page}&per_page={per_page}" if prev_page else None

    if serializer_class:
        # Serialize the paginated results using the provided serializer class
        the_serializer = serializer_class(data=paginated_results, many=True)
        the_serializer.is_valid()
        
        serialized_results = the_serializer.data
    else:
        serialized_results = paginated_results

    if type == 'single':
        # If a single result is expected, handle the case where the results list is empty
        if paginated_results:
            to_index = from_index
            response_data = OrderedDict([(wrap, serialized_results[0])])
        else:
            # If results list is empty, return an empty response
            response_data = OrderedDict([(wrap, {})])
    elif type == 'all':
        if paginated_results:
            to_index = from_index
            response_data = OrderedDict([(wrap, serialized_results)])
        else:
            # If results list is empty, return an empty response
            response_data = OrderedDict([(wrap, [])])
    else:
        # Construct Laravel-style response with pagination information
        response_data = OrderedDict([
            (wrap, serialized_results),
            ('per_page', len(serialized_results)),
            ('current_page', page),
            ('last_page', total_pages),
            ('next_page_url', next_page_url),
            ('prev_page_url', prev_page_url),
            ('total', len(results)),
            ('from', from_index),
            ('to', to_index),
        ])

    return response_data

def exclude_fields(representation, fields_to_exclude):
    for field in list(fields_to_exclude):
        representation.pop(field, None)
    return representation


from functools import wraps
from rest_framework.exceptions import ValidationError

def petra_dto(form_class):
    """
    Decorator that validates incoming request data using a form class.
    Handles both JSON and form-encoded data.
    
    Args:
        form_class: Django Form class to validate the request data
    """
    def decorator(view_func):
        @wraps(view_func)
        def wrapper(self, request, *args, **kwargs):
            if request.content_type == 'application/json':
                form = form_class(request.data)
            else:
                form = form_class(request.POST)

            if form.is_valid():
                return view_func(self, request, form, *args, **kwargs)
            
            # Format the error response
            formatted_errors = {
                'success': False,
                'errors': {}
            }

            for field, error_list in form.errors.items():
                # Convert to array format
                formatted_errors['errors'][field] = [
                    f"The {field} {error_list[0]}"  # Format error message
                ]
            
            raise ValidationError(formatted_errors)
        return wrapper
    return decorator






from celery import shared_task
from django_petra.mailer.send_mail import SendMail

@shared_task()
def petra_mail(email_data):
    mail = SendMail()
    mail.from_email(email_data['from'])
    mail.to(email_data['to'])
    mail.subject(email_data["subject"])
    mail.template(template_name=email_data["template_name"], context=email_data["context"])
    mail.send()



from django_petra.filesystem.the_disks.local import LocalStorage
from django_petra.filesystem.the_disks.s3 import S3Storage

class Storage:
    def __init__(self, disk='local'):
        self.disk = disk
        self.local_storage = LocalStorage()
        self.s3_storage = None
        self.set_disk(disk)
        

    def set_disk(self, storage_type):
        if self.disk == 's3':
            self.s3_storage = S3Storage()
        else:
            self.disk = storage_type

    def put(self, path, contents):
        if self.disk == 'local':
            self.local_storage.put(path, contents)
        elif self.disk == 's3':
            self.s3_storage.put(path, contents)

    def get(self, path):
        if self.disk == 'local':
            return self.local_storage.get(path)
        elif self.disk == 's3':
            return self.s3_storage.get(path)

    def delete(self, path):
        if self.disk == 'local':
            self.local_storage.delete(path)
        elif self.disk == 's3':
            self.s3_storage.delete(path)

    def update(self, path, contents):
        if self.disk == 'local':
            self.local_storage.update(path, contents)
        elif self.disk == 's3':
            self.s3_storage.update(path, contents)

    def exists(self, path):
        if self.disk == 'local':
            return self.local_storage.exists(path)
        elif self.disk == 's3':
            return self.s3_storage.exists(path)

    def download(self, source_path, destination_path):
        if self.disk == 'local':
            self.local_storage.download(source_path, destination_path)
        elif self.disk == 's3':
            self.s3_storage.download(source_path, destination_path)

    def url(self, path):
        if self.disk == 'local':
            return self.local_storage.url(path)
        elif self.disk == 's3':
            return self.s3_storage.url(path)

    def temporary_url(self, path, expiration=3600):
        if self.disk == 'local':
            raise ValueError("Temporary URL is not supported for local storage.")
        elif self.disk == 's3':
            return self.s3_storage.temporary_url(path, expiration)

    def all_files(self, directory):
        if self.disk == 'local':
            return self.local_storage.all_files(directory)
        elif self.disk == 's3':
            return self.s3_storage.all_files(directory)

    def directories(self, directory):
        if self.disk == 'local':
            return self.local_storage.directories(directory)
        elif self.disk == 's3':
            return self.s3_storage.directories(directory)

    def append(self, path, contents):
        if self.disk == 'local':
            self.local_storage.append(path, contents)
        elif self.disk == 's3':
            self.s3_storage.append(path, contents)

    def prepend(self, path, contents):
        if self.disk == 'local':
            self.local_storage.prepend(path, contents)
        elif self.disk == 's3':
            self.s3_storage.prepend(path, contents)

    def copy(self, source, destination):
        if self.disk == 'local':
            self.local_storage.copy(source, destination)
        elif self.disk == 's3':
            self.s3_storage.copy(source, destination)

    def move(self, source, destination):
        if self.disk == 'local':
            self.local_storage.move(source, destination)
        elif self.disk == 's3':
            self.s3_storage.move(source, destination)
    
    def content_type(self, path):
        if self.disk == 'local':
            return self.local_storage.content_type(path)
        elif self.disk == 's3':
            return self.s3_storage.content_type(path)

# Create an instance of the Storage class
storage = Storage()